



celery --version
IF %ERRORLEVEL% NEQ 0 exit /B 1
celery worker --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
celery beat --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
celery multi help
IF %ERRORLEVEL% NEQ 0 exit /B 1
celery shell --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
